==> puyapy-2.1.1-py3-none-any.whl: puya/awst_build/eb/storage/box_ref.py <==
import typing
from collections.abc import Sequence

import mypy.nodes

from puya.awst import wtypes
from puya.awst.nodes import BoxValueExpression, Expression, IntrinsicCall, Not, StateExists
from puya.awst_build import pytypes
from puya.awst_build.eb import _expect as expect
from puya.awst_build.eb._base import FunctionBuilder, NotIterableInstanceExpressionBuilder
from puya.awst_build.eb._bytes_backed import BytesBackedInstanceExpressionBuilder
from puya.awst_build.eb._utils import dummy_value
from puya.awst_build.eb.bool import BoolExpressionBuilder
from puya.awst_build.eb.factories import builder_for_instance
from puya.awst_build.eb.interface import InstanceBuilder, NodeBuilder, TypeBuilder
from puya.awst_build.eb.none import NoneExpressionBuilder
from puya.awst_build.eb.storage._common import BoxGetExpressionBuilder, BoxMaybeExpressionBuilder
from puya.awst_build.eb.storage._storage import StorageProxyDefinitionBuilder, extract_key_override
from puya.awst_build.eb.storage._util import BoxProxyConstructorResult, box_length_checked
from puya.awst_build.eb.uint64 import UInt64ExpressionBuilder
from puya.awst_build.utils import get_arg_mapping
from puya.parse import SourceLocation


class BoxRefTypeBuilder(TypeBuilder[pytypes.StorageProxyType]):
    def __init__(self, location: SourceLocation) -> None:
        super().__init__(pytypes.BoxRefType, location)

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        key_arg_name = "key"
        arg_mapping, _ = get_arg_mapping(
            optional_kw_only=[key_arg_name],
            args=args,
            arg_names=arg_names,
            call_location=location,
            raise_on_missing=False,
        )
        key_arg = arg_mapping.get(key_arg_name)
        key_override = extract_key_override(key_arg, location, typ=wtypes.box_key)
        if key_override is None:
            return StorageProxyDefinitionBuilder(
                self.produces(), location=location, description=None
            )
        return _BoxRefProxyExpressionBuilderFromConstructor(expr=key_override)


class BoxRefProxyExpressionBuilder(
    NotIterableInstanceExpressionBuilder[pytypes.StorageProxyType],
    BytesBackedInstanceExpressionBuilder[pytypes.StorageProxyType],
    bytes_member="key",
):
    def __init__(self, expr: Expression, member_name: str | None = None):
        super().__init__(pytypes.BoxRefType, expr)
        self._member_name = member_name

    def _box_key_expr(self, location: SourceLocation) -> BoxValueExpression:
        if self._member_name:
            exists_assertion_message = f"check self.{self._member_name} exists"
        else:
            exists_assertion_message = "check BoxRef exists"
        return BoxValueExpression(
            key=self.resolve(),
            wtype=wtypes.bytes_wtype,
            exists_assertion_message=exists_assertion_message,
            source_location=location,
        )

    @typing.override
    def bool_eval(self, location: SourceLocation, *, negate: bool = False) -> InstanceBuilder:
        box_exists = StateExists(
            field=self._box_key_expr(location),
            source_location=location,
        )
        return BoolExpressionBuilder(
            Not(expr=box_exists, source_location=location) if negate else box_exists
        )

    @typing.override
    def member_access(
        self, name: str, expr: mypy.nodes.Expression, location: SourceLocation
    ) -> NodeBuilder:
        match name:
            case "create":
                return _Create(location, box_proxy=self.resolve())
            case "delete":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_del",
                    args={},
                    return_type=pytypes.BoolType,
                )
            case "extract":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_extract",
                    args={"start_index": pytypes.UInt64Type, "length": pytypes.UInt64Type},
                    return_type=pytypes.BytesType,
                )
            case "resize":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_resize",
                    args={"new_size": pytypes.UInt64Type},
                    return_type=pytypes.NoneType,
                )
            case "replace":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_replace",
                    args={"start_index": pytypes.UInt64Type, "value": pytypes.BytesType},
                    return_type=pytypes.NoneType,
                )
            case "splice":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_splice",
                    args={
                        "start_index": pytypes.UInt64Type,
                        "length": pytypes.UInt64Type,
                        "value": pytypes.BytesType,
                    },
                    return_type=pytypes.NoneType,
                )

            case "get":
                return BoxGetExpressionBuilder(
                    self._box_key_expr(location), content_type=pytypes.BytesType
                )
            case "put":
                return _Put(location, box_proxy=self.resolve())
            case "maybe":
                return BoxMaybeExpressionBuilder(
                    self._box_key_expr(location), content_type=pytypes.BytesType
                )
            case "length":
                return UInt64ExpressionBuilder(
                    box_length_checked(self._box_key_expr(location), location)
                )
            case _:
                return super().member_access(name, expr, location)


class _BoxRefProxyExpressionBuilderFromConstructor(
    BoxRefProxyExpressionBuilder, BoxProxyConstructorResult
):
    pass


class _IntrinsicMethod(FunctionBuilder):
    def __init__(
        self,
        location: SourceLocation,
        *,
        box_proxy: Expression,
        op_code: str,
        args: dict[str, pytypes.PyType],
        return_type: pytypes.PyType,
    ) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy
        self.op_code = op_code
        self.args = args
        self.return_type = return_type

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        args_map, any_missing = get_arg_mapping(
            required_positional_names=list(self.args.keys()),
            args=args,
            arg_names=arg_names,
            call_location=location,
            raise_on_missing=False,
        )
        if any_missing:
            return dummy_value(self.return_type, location)
        stack_args = [
            expect.argument_of_type_else_dummy(
                args_map[arg_name], arg_type, resolve_literal=True
            ).resolve()
            for arg_name, arg_type in self.args.items()
        ]
        result_expr = IntrinsicCall(
            op_code=self.op_code,
            stack_args=[self.box_proxy, *stack_args],
            wtype=self.return_type.wtype,
            source_location=location,
        )
        return builder_for_instance(self.return_type, result_expr)


class _Create(FunctionBuilder):
    def __init__(self, location: SourceLocation, *, box_proxy: Expression) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.UInt64Type, location, resolve_literal=True
        )
        size = arg.resolve()
        return BoolExpressionBuilder(
            IntrinsicCall(
                op_code="box_create",
                stack_args=[self.box_proxy, size],
                wtype=wtypes.bool_wtype,
                source_location=location,
            )
        )


class _Put(FunctionBuilder):
    def __init__(self, location: SourceLocation, *, box_proxy: Expression) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.BytesType, location, resolve_literal=True
        )
        data = arg.resolve()
        return NoneExpressionBuilder(
            IntrinsicCall(
                op_code="box_put",
                stack_args=[self.box_proxy, data],
                wtype=wtypes.void_wtype,
                source_location=location,
            )
        )

==> puyapy-2.3.0-py3-none-any.whl: puya/awst_build/eb/storage/box_ref.py <==
import typing
from collections.abc import Sequence

import mypy.nodes

from puya.awst import wtypes
from puya.awst.nodes import BoxValueExpression, Expression, IntrinsicCall, Not, StateExists
from puya.awst_build import pytypes
from puya.awst_build.eb import _expect as expect
from puya.awst_build.eb._base import FunctionBuilder, NotIterableInstanceExpressionBuilder
from puya.awst_build.eb._bytes_backed import BytesBackedInstanceExpressionBuilder
from puya.awst_build.eb._utils import dummy_value
from puya.awst_build.eb.bool import BoolExpressionBuilder
from puya.awst_build.eb.factories import builder_for_instance
from puya.awst_build.eb.interface import InstanceBuilder, NodeBuilder, TypeBuilder
from puya.awst_build.eb.none import NoneExpressionBuilder
from puya.awst_build.eb.storage._common import BoxGetExpressionBuilder, BoxMaybeExpressionBuilder
from puya.awst_build.eb.storage._storage import StorageProxyDefinitionBuilder, extract_key_override
from puya.awst_build.eb.storage._util import BoxProxyConstructorResult, box_length_checked
from puya.awst_build.eb.uint64 import UInt64ExpressionBuilder
from puya.awst_build.utils import get_arg_mapping
from puya.parse import SourceLocation


class BoxRefTypeBuilder(TypeBuilder[pytypes.StorageProxyType]):
    def __init__(self, location: SourceLocation) -> None:
        super().__init__(pytypes.BoxRefType, location)

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        key_arg_name = "key"
        arg_mapping, _ = get_arg_mapping(
            optional_kw_only=[key_arg_name],
            args=args,
            arg_names=arg_names,
            call_location=location,
            raise_on_missing=False,
        )
        key_arg = arg_mapping.get(key_arg_name)
        key_override = extract_key_override(key_arg, location, typ=wtypes.box_key)
        if key_override is None:
            return StorageProxyDefinitionBuilder(
                self.produces(), location=location, description=None
            )
        return _BoxRefProxyExpressionBuilderFromConstructor(expr=key_override)


class BoxRefProxyExpressionBuilder(
    NotIterableInstanceExpressionBuilder[pytypes.StorageProxyType],
    BytesBackedInstanceExpressionBuilder[pytypes.StorageProxyType],
    bytes_member="key",
):
    def __init__(self, expr: Expression, member_name: str | None = None):
        super().__init__(pytypes.BoxRefType, expr)
        self._member_name = member_name

    def _box_key_expr(self, location: SourceLocation) -> BoxValueExpression:
        if self._member_name:
            exists_assertion_message = f"check self.{self._member_name} exists"
        else:
            exists_assertion_message = "check BoxRef exists"
        return BoxValueExpression(
            key=self.resolve(),
            wtype=wtypes.bytes_wtype,
            exists_assertion_message=exists_assertion_message,
            source_location=location,
        )

    @typing.override
    def bool_eval(self, location: SourceLocation, *, negate: bool = False) -> InstanceBuilder:
        box_exists = StateExists(
            field=self._box_key_expr(location),
            source_location=location,
        )
        return BoolExpressionBuilder(
            Not(expr=box_exists, source_location=location) if negate else box_exists
        )

    @typing.override
    def member_access(self, name: str, location: SourceLocation) -> NodeBuilder:
        match name:
            case "create":
                return _Create(location, box_proxy=self.resolve())
            case "delete":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_del",
                    args={},
                    return_type=pytypes.BoolType,
                )
            case "extract":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_extract",
                    args={"start_index": pytypes.UInt64Type, "length": pytypes.UInt64Type},
                    return_type=pytypes.BytesType,
                )
            case "resize":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_resize",
                    args={"new_size": pytypes.UInt64Type},
                    return_type=pytypes.NoneType,
                )
            case "replace":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_replace",
                    args={"start_index": pytypes.UInt64Type, "value": pytypes.BytesType},
                    return_type=pytypes.NoneType,
                )
            case "splice":
                return _IntrinsicMethod(
                    location,
                    box_proxy=self.resolve(),
                    op_code="box_splice",
                    args={
                        "start_index": pytypes.UInt64Type,
                        "length": pytypes.UInt64Type,
                        "value": pytypes.BytesType,
                    },
                    return_type=pytypes.NoneType,
                )

            case "get":
                return BoxGetExpressionBuilder(
                    self._box_key_expr(location), content_type=pytypes.BytesType
                )
            case "put":
                return _Put(location, box_proxy=self.resolve())
            case "maybe":
                return BoxMaybeExpressionBuilder(
                    self._box_key_expr(location), content_type=pytypes.BytesType
                )
            case "length":
                return UInt64ExpressionBuilder(
                    box_length_checked(self._box_key_expr(location), location)
                )
            case _:
                return super().member_access(name, location)


class _BoxRefProxyExpressionBuilderFromConstructor(
    BoxRefProxyExpressionBuilder, BoxProxyConstructorResult
):
    pass


class _IntrinsicMethod(FunctionBuilder):
    def __init__(
        self,
        location: SourceLocation,
        *,
        box_proxy: Expression,
        op_code: str,
        args: dict[str, pytypes.PyType],
        return_type: pytypes.PyType,
    ) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy
        self.op_code = op_code
        self.args = args
        self.return_type = return_type

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        args_map, any_missing = get_arg_mapping(
            required_positional_names=list(self.args.keys()),
            args=args,
            arg_names=arg_names,
            call_location=location,
            raise_on_missing=False,
        )
        if any_missing:
            return dummy_value(self.return_type, location)
        stack_args = [
            expect.argument_of_type_else_dummy(
                args_map[arg_name], arg_type, resolve_literal=True
            ).resolve()
            for arg_name, arg_type in self.args.items()
        ]
        result_expr = IntrinsicCall(
            op_code=self.op_code,
            stack_args=[self.box_proxy, *stack_args],
            wtype=self.return_type.wtype,
            source_location=location,
        )
        return builder_for_instance(self.return_type, result_expr)


class _Create(FunctionBuilder):
    def __init__(self, location: SourceLocation, *, box_proxy: Expression) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.UInt64Type, location, resolve_literal=True
        )
        size = arg.resolve()
        return BoolExpressionBuilder(
            IntrinsicCall(
                op_code="box_create",
                stack_args=[self.box_proxy, size],
                wtype=wtypes.bool_wtype,
                source_location=location,
            )
        )


class _Put(FunctionBuilder):
    def __init__(self, location: SourceLocation, *, box_proxy: Expression) -> None:
        super().__init__(location)
        self.box_proxy = box_proxy

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.BytesType, location, resolve_literal=True
        )
        data = arg.resolve()
        return NoneExpressionBuilder(
            IntrinsicCall(
                op_code="box_put",
                stack_args=[self.box_proxy, data],
                wtype=wtypes.void_wtype,
                source_location=location,
            )
        )
